# Copyright 2011-2014 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""Basic script utilities."""

__metaclass__ = type
__all__ = [
    'ScriptBase',
    'ScriptFailure',
    ]

import errno
import fcntl
import logging
from optparse import OptionParser
import os
import re
import sys
import time

from launchpadlib.launchpad import Launchpad


LOCK_PATH = "/var/lock/"


class ScriptFormatter(logging.Formatter):
    """logging.Formatter encoding our preferred output format."""

    def __init__(self, fmt=None, datefmt="%Y-%m-%d %H:%M:%S"):
        if fmt is None:
            fmt = '%(asctime)s %(levelname)-7s %(message)s'
        logging.Formatter.__init__(self, fmt, datefmt)
        # Output should be UTC.
        self.converter = time.gmtime


def log_unhandled_exception_and_exit(func):
    """Decorator that logs unhandled exceptions via the logging module.

    Exceptions are reraised except at the top level. ie. exceptions are
    only propagated to the outermost decorated method. At the top level,
    an exception causes the script to terminate.
    """

    def log_unhandled_exceptions_func(self, *args, **kw):
        try:
            self._log_unhandled_exceptions_level += 1
            return func(self, *args, **kw)
        except Exception:
            if self._log_unhandled_exceptions_level == 1:
                # self.logger is setup in GenerateExtraOverrides.__init__()
                # so we can use it.
                self.logger.exception("Unhandled exception")
                sys.exit(1)
            else:
                raise
        finally:
            self._log_unhandled_exceptions_level -= 1
    return log_unhandled_exceptions_func


class ScriptFailure(Exception):
    """Something bad happened and the script is going away."""


class ScriptBase:
    """A basic script class."""

    # State for the log_unhandled_exceptions_and_exit decorator.
    _log_unhandled_exceptions_level = 0

    def __init__(self, name=None, test_args=None):
        if name is None:
            self._name = self.__class__.__name__.lower()
        else:
            self._name = name
        self.parser = OptionParser()
        self.parser.add_option(
            "-l", "--launchpad", dest="launchpad_instance",
            default="production")
        self.add_my_options()
        self.options, self.args = self.parser.parse_args(args=test_args)

        self.logger = logging.getLogger()
        hdlr = logging.StreamHandler()
        self.logger.setLevel(0)
        hdlr.setLevel(logging.DEBUG)
        hdlr.setFormatter(ScriptFormatter())
        self.logger.addHandler(hdlr)

    @property
    def name(self):
        return self._name

    def main(self):
        raise NotImplementedError

    def add_my_options(self):
        pass

    def login(self):
        return Launchpad.login_anonymously(
            self.name, self.options.launchpad_instance)

    @property
    def lockfilename(self):
        return "ubuntu-archive-publishing-%s.lock" % self.name

    @property
    def lockfilepath(self):
        return os.path.join(LOCK_PATH, self.lockfilename)

    def setup_lock(self):
        """Create lockfile.

        Note that this will create a lockfile even if you don't actually
        use it.
        """
        if os.path.exists(self.lockfilepath):
            self.previous_lockfile_present = True
        else:
            self.previous_lockfile_present = False
        self.flock = open(self.lockfilepath, "w")
        self.fdlock = self.flock.fileno()

    @log_unhandled_exception_and_exit
    def lock_or_die(self):
        """Attempt to lock, and sys.exit(1) if the lock's already taken."""
        self.setup_lock()
        self.logger.info("Creating lockfile: %s", self.lockfilepath)
        try:
            fcntl.flock(self.fdlock, fcntl.LOCK_EX | fcntl.LOCK_NB)
        except IOError as message:
            match = re.search("^\[Errno ([0-9]+)\]", str(message))
            if match:
                if int(match.group(1)) == errno.EWOULDBLOCK:
                    self.logger.info('Lockfile %s in use' % self.lockfilepath)
                    sys.exit(1)
                else:
                    raise Exception(
                        'Cannot acquire lock on file "%s": %s' %
                        (self.lockfilepath, message))
            else:
                raise Exception('Malformed error message "%s"' % message)
        if self.previous_lockfile_present:
            self.logger.warn("Stale lockfile detected and claimed.")
        self.is_locked = True

    @log_unhandled_exception_and_exit
    def unlock(self):
        """Release the lock. Do this before going home."""
        if not self.is_locked:
            return
        self.logger.debug('Removing lock file: %s', self.lockfilepath)
        os.unlink(self.lockfilepath)
        try:
            fcntl.flock(self.fdlock, fcntl.LOCK_UN)
        except IOError:
            raise Exception('Unlock of file "%s" failed' % self.lockfilepath)
        self.is_locked = False

    @log_unhandled_exception_and_exit
    def run(self):
        """Actually run the script."""
        try:
            self.main()
        except ScriptFailure as e:
            self.logger.error(str(e))
            sys.exit(1)

    @log_unhandled_exception_and_exit
    def lock_and_run(self):
        """Call lock_or_die(), and then run() the script.

        Will die with sys.exit(1) if the locking call fails.
        """
        self.lock_or_die()
        try:
            self.run()
        finally:
            self.unlock()
