# Copyright 2011-2014 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""Test for the `generate-extra-overrides` script."""

from __future__ import print_function

__metaclass__ = type

from collections import defaultdict
from functools import partial
import gzip
import logging
from optparse import OptionValueError
import os
try:
    from unittest import mock
except ImportError:
    import mock

from germinate import (
    archive,
    germinator,
    seeds,
    )

from scripts.generate_extra_overrides import (
    AtomicFile,
    file_exists,
    GenerateExtraOverrides,
    ScriptFailure,
    )
from tests.testing import (
    ensure_directory_exists,
    factory,
    file_contents,
    open_for_writing,
    run_script,
    TestCase,
    write_file,
    )


class TestAtomicFile(TestCase):
    """Tests for the AtomicFile helper class."""

    def test_atomic_file_creates_file(self):
        # AtomicFile creates the named file with the requested contents.
        self.useTempDir()
        filename = factory.getUniqueString()
        text = factory.getUniqueString()
        with AtomicFile(filename) as test:
            test.write(text)
        self.assertEqual(text, file_contents(filename))

    def test_atomic_file_removes_dot_new(self):
        # AtomicFile does not leave .new files lying around.
        self.useTempDir()
        filename = factory.getUniqueString()
        with AtomicFile(filename):
            pass
        self.assertFalse(file_exists("%s.new" % filename))


class IndexStanzaFields:
    """Store and format ordered Index Stanza fields."""

    def __init__(self):
        self._names_lower = set()
        self.fields = []

    def append(self, name, value):
        """Append an (field, value) tuple to the internal list.

        Then we can use the FIFO-like behaviour in makeOutput().
        """
        if name.lower() in self._names_lower:
            return
        self._names_lower.add(name.lower())
        self.fields.append((name, value))

    def extend(self, entries):
        """Extend the internal list with the key-value pairs in entries.
        """
        for name, value in entries:
            self.append(name, value)

    def makeOutput(self):
        """Return a line-by-line aggregation of appended fields.

        Empty fields values will cause the exclusion of the field.
        The output order will preserve the insertion order, FIFO.
        """
        output_lines = []
        for name, value in self.fields:
            if not value:
                continue

            # Do not add separation space for the special file list fields.
            if name not in ('Files', 'Checksums-Sha1', 'Checksums-Sha256'):
                value = ' %s' % value

            output_lines.append('%s:%s' % (name, value))

        return '\n'.join(output_lines)


class MockArchivePublisherBase(mock.MagicMock):

    def getIndexStanza(self):
        fields = self.buildIndexStanzaFields()
        return fields.makeOutput()


class MockSourcePackagePublishingHistory(MockArchivePublisherBase):
    """A mock source package publishing history.

    This is a crude amalgam of SourcePackageRelease and
    SourcePackagePublishingHistory, but it'll do for us.
    """

    def __init__(self, sourcepackagename=None):
        super(MockSourcePackagePublishingHistory, self).__init__()
        if sourcepackagename is None:
            sourcepackagename = factory.getUniqueString()
        self.sourcepackagename = sourcepackagename
        self.binaries = []
        self.version = unicode(factory.getUniqueInteger()) + "version"
        self.section = factory.getUniqueString(prefix="section")
        self.dsc_maintainer_rfc822 = "%s <%s>" % (
            factory.getUniqueString(), factory.getUniqueEmailAddress())
        self.architecturehintlist = "all"
        self.dsc_format = "1.0"

    @property
    def dsc_binaries(self):
        return ", ".join(self.binaries)

    def buildIndexStanzaFields(self):
        fields = IndexStanzaFields()
        fields.append("Package", self.sourcepackagename)
        fields.append("Binary", self.dsc_binaries)
        fields.append("Version", self.version)
        fields.append("Section", self.section)
        fields.append("Maintainer", self.dsc_maintainer_rfc822)
        fields.append("Architecture", self.architecturehintlist)
        fields.append("Format", self.dsc_format)
        return fields


class MockBinaryPackagePublishingHistory(MockArchivePublisherBase):
    """A mock binary package publishing history.

    This is a crude amalgam of BinaryPackageRelease and
    BinaryPackagePublishingHistory, but it'll do for us.
    """

    def __init__(self, spph, das, binarypackagename=None, depends=None):
        super(MockBinaryPackagePublishingHistory, self).__init__()
        if binarypackagename is None:
            binarypackagename = factory.getUniqueString()
        self.spph = spph
        self.das = das
        self.binarypackagename = binarypackagename
        self.priority = "optional"
        self.depends = depends

    def buildIndexStanzaFields(self):
        fields = IndexStanzaFields()
        fields.append("Package", self.binarypackagename)
        fields.append("Source", self.spph.sourcepackagename)
        fields.append("Priority", self.priority)
        fields.append("Section", self.spph.section)
        fields.append("Maintainer", self.spph.dsc_maintainer_rfc822)
        fields.append("Architecture", self.das.architecture_tag)
        fields.append("Version", self.spph.version)
        fields.append("Depends", self.depends)
        return fields


class MockDistroArchSeries(mock.MagicMock):
    def __init__(self, distroseries, architecture_tag=None):
        super(MockDistroArchSeries, self).__init__()
        if architecture_tag is None:
            architecture_tag = factory.getUniqueString(prefix="arch")
        self.distroseries = distroseries
        self.architecture_tag = architecture_tag
        self.test_binaries = defaultdict(list)


class MockDistroSeries(mock.MagicMock):
    def __init__(self, distribution, status="Active Development"):
        super(MockDistroSeries, self).__init__()
        self.name = factory.getUniqueString(prefix="distroseries")
        self.distribution = distribution
        self.component_names = []
        self.status = status
        self.architectures = []
        self.test_sources = defaultdict(list)

    def makeDistroArchSeries(self, enabled=True, *args, **kwargs):
        das = MockDistroArchSeries(distroseries=self, *args, **kwargs)
        if enabled:
            self.architectures.append(das)
        return das


class MockDistribution(mock.MagicMock):
    def __init__(self, name):
        super(MockDistribution, self).__init__()
        self.name = name
        self.series = []

    def makeDistroSeries(self, *args, **kwargs):
        series = MockDistroSeries(distribution=self, *args, **kwargs)
        self.series.append(series)
        return series


class MockLaunchpad(mock.MagicMock):
    def __init__(self):
        super(MockLaunchpad, self).__init__(name="Launchpad")
        self.distributions = {}

    def makeDistribution(self, name=None):
        if name is None:
            name = factory.getUniqueString()
        distribution = MockDistribution(name)
        self.distributions[name] = distribution
        return distribution


class TestGenerateExtraOverrides(TestCase):
    """Tests for the actual `GenerateExtraOverrides` script."""

    def setUp(self):
        super(TestGenerateExtraOverrides, self).setUp()
        self.seeddir = self.makeTemporaryDirectory()
        self.launchpad = MockLaunchpad()
        # XXX cjwatson 2011-12-06 bug=694140: Make sure germinate doesn't
        # lose its loggers between tests, due to Launchpad's messing with
        # global log state.
        archive._logger = logging.getLogger("germinate.archive")
        germinator._logger = logging.getLogger("germinate.germinator")
        seeds._logger = logging.getLogger("germinate.seeds")

    def assertFilesEqual(self, expected_path, observed_path):
        self.assertEqual(
            file_contents(expected_path), file_contents(observed_path))

    def makeDistro(self):
        """Create a distribution for testing.

        The distribution will have a root directory set up, which will
        be cleaned up after the test.  It will have an attached archive.
        """
        publish_root_dir = self.makeTemporaryDirectory()
        os.environ["ARCHIVEROOTS"] = publish_root_dir
        return self.launchpad.makeDistribution()

    def makeScript(self, distribution, run_setup=True, extra_args=None):
        """Create a script for testing."""
        test_args = []
        if distribution is not None:
            test_args.extend(["-d", distribution.name])
        if extra_args is not None:
            test_args.extend(extra_args)
        script = GenerateExtraOverrides(test_args=test_args)
        for handler in list(script.logger.handlers):
            script.logger.removeHandler(handler)
        script.logger.addHandler(logging.NullHandler())
        script.login = lambda: self.launchpad
        if distribution is not None and run_setup:
            script.setUp()
        else:
            script.distribution = distribution
        return script

    def setUpDistroAndScript(self, series_statuses=["Active Development"],
                             **kwargs):
        """Helper wrapping distro and script setup."""
        self.distro = self.makeDistro()
        self.distroseries = [
            self.distro.makeDistroSeries(status=status)
            for status in series_statuses]
        self.script = self.makeScript(self.distro, **kwargs)

    def setUpComponent(self, component=None):
        """Create a component and attach it to all distroseries."""
        if component is None:
            component = factory.getUniqueString()
        for distroseries in self.distroseries:
            distroseries.component_names.append(component)
        return component

    def makePackage(self, component, dases, **kwargs):
        """Create a published source and binary package for testing."""
        package_name = factory.getUniqueString()
        spph = MockSourcePackagePublishingHistory(
            sourcepackagename=package_name)
        dases[0].distroseries.test_sources[component].append(spph)
        for das in dases:
            bpph = MockBinaryPackagePublishingHistory(
                spph, das, binarypackagename=package_name, **kwargs)
            das.test_binaries[component].append(bpph)
        return package_name

    def makeIndexFiles(self, script, distroseries):
        """Create a limited subset of index files for testing."""
        for component in distroseries.component_names:
            sources_path = os.path.join(
                script.archiveroot, "dists", distroseries.name, component,
                "source", "Sources.gz")
            ensure_directory_exists(os.path.dirname(sources_path))
            sources_index = gzip.GzipFile(sources_path, "wb")
            for spp in distroseries.test_sources[component]:
                stanza = spp.getIndexStanza().encode("utf-8") + "\n\n"
                sources_index.write(stanza)
            sources_index.close()

            for arch in distroseries.architectures:
                packages_path = os.path.join(
                    script.archiveroot, "dists", distroseries.name, component,
                    "binary-%s" % arch.architecture_tag, "Packages.gz")
                ensure_directory_exists(os.path.dirname(packages_path))
                packages_index = gzip.GzipFile(packages_path, "wb")
                for bpp in arch.test_binaries[component]:
                    stanza = bpp.getIndexStanza().encode("utf-8") + "\n\n"
                    packages_index.write(stanza)
                packages_index.close()

    def composeSeedPath(self, flavour, series_name, seed_name):
        return os.path.join(
            self.seeddir, "%s.%s" % (flavour, series_name), seed_name)

    def makeSeedStructure(self, flavour, series_name, seed_names,
                          seed_inherit={}):
        """Create a simple seed structure file."""
        structure_path = self.composeSeedPath(
            flavour, series_name, "STRUCTURE")
        with open_for_writing(structure_path, "w") as structure:
            for seed_name in seed_names:
                inherit = seed_inherit.get(seed_name, [])
                line = "%s: %s" % (seed_name, " ".join(inherit))
                print(line.strip(), file=structure)

    def makeSeed(self, flavour, series_name, seed_name, entries,
                 headers=None):
        """Create a simple seed file."""
        seed_path = self.composeSeedPath(flavour, series_name, seed_name)
        with open_for_writing(seed_path, "w") as seed:
            if headers is not None:
                for header in headers:
                    print(header, file=seed)
                print(file=seed)
            for entry in entries:
                print(" * %s" % entry, file=seed)

    def getTaskNameFromSeed(self, script, flavour, series_name, seed,
                            primary_flavour):
        """Use script to parse a seed and return its task name."""
        seed_path = self.composeSeedPath(flavour, series_name, seed)
        with open(seed_path) as seed_text:
            task_headers = script.parseTaskHeaders(seed_text)
        return script.getTaskName(
            task_headers, flavour, seed, primary_flavour)

    def getTaskSeedsFromSeed(self, script, flavour, series_name, seed):
        """Use script to parse a seed and return its task seed list."""
        seed_path = self.composeSeedPath(flavour, series_name, seed)
        with open(seed_path) as seed_text:
            task_headers = script.parseTaskHeaders(seed_text)
        return script.getTaskSeeds(task_headers, seed)

    def test_name_is_consistent(self):
        # Script instances for the same distro get the same name.
        distro = self.launchpad.makeDistribution()
        self.assertEqual(
            GenerateExtraOverrides(test_args=["-d", distro.name]).name,
            GenerateExtraOverrides(test_args=["-d", distro.name]).name)

    def test_name_is_unique_for_each_distro(self):
        # Script instances for different distros get different names.
        self.assertNotEqual(
            GenerateExtraOverrides(
                test_args=["-d", self.launchpad.makeDistribution().name]).name,
            GenerateExtraOverrides(
                test_args=["-d", self.launchpad.makeDistribution().name]).name)

    def test_requires_distro(self):
        # The --distribution or -d argument is mandatory.
        script = self.makeScript(None)
        self.assertRaises(OptionValueError, script.processOptions)

    def test_requires_real_distro(self):
        # An incorrect distribution name is flagged as an invalid option
        # value.
        script = self.makeScript(
            None, extra_args=["-d", factory.getUniqueString()])
        self.assertRaises(OptionValueError, script.processOptions)

    def test_looks_up_distro(self):
        # The script looks up and keeps the distribution named on the
        # command line.
        self.setUpDistroAndScript()
        self.assertEqual(self.distro, self.script.distribution)

    def test_prefers_development_distro_series(self):
        # The script prefers a DEVELOPMENT series for the named
        # distribution over CURRENT and SUPPORTED series.
        self.setUpDistroAndScript(
            ["Supported", "Current Stable Release", "Active Development"])
        self.assertEqual([self.distroseries[2]], self.script.series)

    def test_permits_frozen_distro_series(self):
        # If there is no DEVELOPMENT series, a FROZEN one will do.
        self.setUpDistroAndScript(
            ["Supported", "Current Stable Release", "Pre-release Freeze"])
        self.assertEqual([self.distroseries[2]], self.script.series)

    def test_requires_development_frozen_distro_series(self):
        # If there is no DEVELOPMENT or FROZEN series, the script fails.
        self.setUpDistroAndScript(
            ["Supported", "Current Stable Release"], run_setup=False)
        self.assertRaises(ScriptFailure, self.script.processOptions)

    def test_multiple_development_frozen_distro_series(self):
        # If there are multiple DEVELOPMENT or FROZEN series, they are all
        # used.
        self.setUpDistroAndScript([
            "Active Development", "Active Development",
            "Pre-release Freeze", "Pre-release Freeze",
            ])
        self.assertCountEqual(self.distroseries, self.script.series)

    def test_components_exclude_partner(self):
        # If a 'partner' component exists, it is excluded.
        self.setUpDistroAndScript()
        self.setUpComponent(component="main")
        self.setUpComponent(component="partner")
        self.assertEqual(1, len(self.script.series))
        self.assertEqual(
            ["main"], self.script.getComponents(self.script.series[0]))

    def test_compose_output_path_in_germinateroot(self):
        # Output files are written to the correct locations under
        # germinateroot.
        self.setUpDistroAndScript()
        flavour = factory.getUniqueString()
        arch = factory.getUniqueString()
        base = factory.getUniqueString()
        output = self.script.composeOutputPath(
            flavour, self.distroseries[0].name, arch, base)
        self.assertEqual(
            "%s/%s_%s_%s_%s" % (
                self.script.germinateroot, base, flavour,
                self.distroseries[0].name, arch),
            output)

    def test_make_seed_structures_missing_seeds(self):
        # makeSeedStructures ignores missing seeds.
        self.setUpDistroAndScript()
        series_name = self.distroseries[0].name
        flavour = factory.getUniqueString()

        structures = self.script.makeSeedStructures(
            series_name, [flavour], seed_bases=["file://%s" % self.seeddir])
        self.assertEqual({}, structures)

    def test_make_seed_structures_empty_seed_structure(self):
        # makeSeedStructures ignores an empty seed structure.
        self.setUpDistroAndScript()
        series_name = self.distroseries[0].name
        flavour = factory.getUniqueString()
        self.makeSeedStructure(flavour, series_name, [])

        structures = self.script.makeSeedStructures(
            series_name, [flavour], seed_bases=["file://%s" % self.seeddir])
        self.assertEqual({}, structures)

    def test_make_seed_structures_valid_seeds(self):
        # makeSeedStructures reads valid seeds successfully.
        self.setUpDistroAndScript()
        series_name = self.distroseries[0].name
        flavour = factory.getUniqueString()
        seed = factory.getUniqueString()
        self.makeSeedStructure(flavour, series_name, [seed])
        self.makeSeed(flavour, series_name, seed, [])

        structures = self.script.makeSeedStructures(
            series_name, [flavour], seed_bases=["file://%s" % self.seeddir])
        self.assertIn(flavour, structures)

    def fetchGerminatedOverrides(self, script, distroseries, arch, flavours):
        """Helper to call script.germinateArch and return overrides."""
        structures = script.makeSeedStructures(
            distroseries.name, flavours,
            seed_bases=["file://%s" % self.seeddir])

        _, overrides, _ = script.germinateArch(
            distroseries.name, script.getComponents(distroseries), arch,
            flavours, structures)
        return overrides.splitlines()

    def test_germinate_output(self):
        # A single call to germinateArch produces output for all flavours on
        # one architecture.
        self.setUpDistroAndScript()
        series_name = self.distroseries[0].name
        component = self.setUpComponent()
        das = self.distroseries[0].makeDistroArchSeries()
        arch = das.architecture_tag
        one = self.makePackage(component, [das])
        two = self.makePackage(component, [das])
        self.makeIndexFiles(self.script, self.distroseries[0])

        flavour_one = factory.getUniqueString()
        flavour_two = factory.getUniqueString()
        seed = factory.getUniqueString()
        self.makeSeedStructure(flavour_one, series_name, [seed])
        self.makeSeed(flavour_one, series_name, seed, [one])
        self.makeSeedStructure(flavour_two, series_name, [seed])
        self.makeSeed(flavour_two, series_name, seed, [two])

        overrides = self.fetchGerminatedOverrides(
            self.script, self.distroseries[0], arch,
            [flavour_one, flavour_two])
        self.assertEqual([], overrides)

        seed_dir_one = os.path.join(
            self.seeddir, "%s.%s" % (flavour_one, series_name))
        self.assertFilesEqual(
            os.path.join(seed_dir_one, "STRUCTURE"),
            self.script.composeOutputPath(
                flavour_one, series_name, arch, "structure"))
        self.assertTrue(file_exists(self.script.composeOutputPath(
            flavour_one, series_name, arch, "all")))
        self.assertTrue(file_exists(self.script.composeOutputPath(
            flavour_one, series_name, arch, "all.sources")))
        self.assertTrue(file_exists(self.script.composeOutputPath(
            flavour_one, series_name, arch, seed)))

        seed_dir_two = os.path.join(
            self.seeddir, "%s.%s" % (flavour_two, series_name))
        self.assertFilesEqual(
            os.path.join(seed_dir_two, "STRUCTURE"),
            self.script.composeOutputPath(
                flavour_two, series_name, arch, "structure"))
        self.assertTrue(file_exists(self.script.composeOutputPath(
            flavour_two, series_name, arch, "all")))
        self.assertTrue(file_exists(self.script.composeOutputPath(
            flavour_two, series_name, arch, "all.sources")))
        self.assertTrue(file_exists(self.script.composeOutputPath(
            flavour_two, series_name, arch, seed)))

    def test_germinate_output_task(self):
        # germinateArch produces Task extra overrides.
        self.setUpDistroAndScript()
        series_name = self.distroseries[0].name
        component = self.setUpComponent()
        das = self.distroseries[0].makeDistroArchSeries()
        arch = das.architecture_tag
        one = self.makePackage(component, [das])
        two = self.makePackage(component, [das], depends=one)
        three = self.makePackage(component, [das])
        self.makePackage(component, [das])
        self.makeIndexFiles(self.script, self.distroseries[0])

        flavour = factory.getUniqueString()
        seed_one = factory.getUniqueString()
        seed_two = factory.getUniqueString()
        self.makeSeedStructure(flavour, series_name, [seed_one, seed_two])
        self.makeSeed(
            flavour, series_name, seed_one, [two],
            headers=["Task-Description: one"])
        self.makeSeed(
            flavour, series_name, seed_two, [three],
            headers=["Task-Description: two"])

        overrides = self.fetchGerminatedOverrides(
            self.script, self.distroseries[0], arch, [flavour])
        expected_overrides = [
            "%s/%s  Task  %s" % (one, arch, seed_one),
            "%s/%s  Task  %s" % (two, arch, seed_one),
            "%s/%s  Task  %s" % (three, arch, seed_two),
            ]
        self.assertCountEqual(expected_overrides, overrides)

    def test_task_name(self):
        # The Task-Name field is honoured.
        series_name = factory.getUniqueString()
        package = factory.getUniqueString()
        script = self.makeScript(None)

        flavour = factory.getUniqueString()
        seed = factory.getUniqueString()
        task = factory.getUniqueString()
        self.makeSeed(
            flavour, series_name, seed, [package],
            headers=["Task-Name: %s" % task])

        observed_task = self.getTaskNameFromSeed(
            script, flavour, series_name, seed, True)
        self.assertEqual(task, observed_task)

    def test_task_per_derivative(self):
        # The Task-Per-Derivative field is honoured.
        series_name = factory.getUniqueString()
        package = factory.getUniqueString()
        script = self.makeScript(None)

        flavour_one = factory.getUniqueString()
        flavour_two = factory.getUniqueString()
        seed_one = factory.getUniqueString()
        seed_two = factory.getUniqueString()
        self.makeSeed(
            flavour_one, series_name, seed_one, [package],
            headers=["Task-Description: one"])
        self.makeSeed(
            flavour_one, series_name, seed_two, [package],
            headers=["Task-Per-Derivative: 1"])
        self.makeSeed(
            flavour_two, series_name, seed_one, [package],
            headers=["Task-Description: one"])
        self.makeSeed(
            flavour_two, series_name, seed_two, [package],
            headers=["Task-Per-Derivative: 1"])

        observed_task_one_one = self.getTaskNameFromSeed(
            script, flavour_one, series_name, seed_one, True)
        observed_task_one_two = self.getTaskNameFromSeed(
            script, flavour_one, series_name, seed_two, True)
        observed_task_two_one = self.getTaskNameFromSeed(
            script, flavour_two, series_name, seed_one, False)
        observed_task_two_two = self.getTaskNameFromSeed(
            script, flavour_two, series_name, seed_two, False)

        # seed_one is not per-derivative, so it is honoured only for
        # flavour_one and has a global name.
        self.assertEqual(seed_one, observed_task_one_one)
        self.assertIsNone(observed_task_two_one)

        # seed_two is per-derivative, so it is honoured for both flavours
        # and has the flavour name prefixed.
        self.assertEqual(
            "%s-%s" % (flavour_one, seed_two), observed_task_one_two)
        self.assertEqual(
            "%s-%s" % (flavour_two, seed_two), observed_task_two_two)

    def test_task_seeds(self):
        # The Task-Seeds field is honoured.
        series_name = factory.getUniqueString()
        one = factory.getUniqueString()
        two = factory.getUniqueString()
        script = self.makeScript(None)

        flavour = factory.getUniqueString()
        seed_one = factory.getUniqueString()
        seed_two = factory.getUniqueString()
        self.makeSeed(flavour, series_name, seed_one, [one])
        self.makeSeed(
            flavour, series_name, seed_two, [two],
            headers=["Task-Seeds: %s" % seed_one])

        task_seeds = self.getTaskSeedsFromSeed(
            script, flavour, series_name, seed_two)
        self.assertCountEqual([seed_one, seed_two], task_seeds)

    def test_germinate_output_build_essential(self):
        # germinateArch produces Build-Essential extra overrides.
        self.setUpDistroAndScript()
        series_name = self.distroseries[0].name
        component = self.setUpComponent()
        das = self.distroseries[0].makeDistroArchSeries()
        arch = das.architecture_tag
        package = self.makePackage(component, [das])
        self.makeIndexFiles(self.script, self.distroseries[0])

        flavour = factory.getUniqueString()
        seed = "build-essential"
        self.makeSeedStructure(flavour, series_name, [seed])
        self.makeSeed(flavour, series_name, seed, [package])

        overrides = self.fetchGerminatedOverrides(
            self.script, self.distroseries[0], arch, [flavour])
        self.assertCountEqual(
            ["%s/%s  Build-Essential  yes" % (package, arch)], overrides)

    def test_removes_only_stale_files(self):
        # removeStaleOutputs removes only stale germinate output files.
        self.setUpDistroAndScript()
        series_name = self.distroseries[0].name
        seed_old_file = "old_flavour_%s_i386" % series_name
        seed_new_file = "new_flavour_%s_i386" % series_name
        other_file = "other-file"
        output = partial(os.path.join, self.script.germinateroot)
        for base in (seed_old_file, seed_new_file, other_file):
            write_file(output(base), "")
        self.script.removeStaleOutputs(series_name, set([seed_new_file]))
        self.assertFalse(os.path.exists(output(seed_old_file)))
        self.assertTrue(os.path.exists(output(seed_new_file)))
        self.assertTrue(os.path.exists(output(other_file)))

    def test_process_missing_seeds(self):
        # The script ignores series with no seed structures.
        flavour = factory.getUniqueString()
        self.setUpDistroAndScript(
            ["Active Development", "Active Development"], extra_args=[flavour])
        self.setUpComponent()
        self.distroseries[0].makeDistroArchSeries()
        self.distroseries[1].makeDistroArchSeries()
        self.makeIndexFiles(self.script, self.distroseries[1])
        seed = factory.getUniqueString()
        self.makeSeedStructure(flavour, self.distroseries[1].name, [seed])
        self.makeSeed(flavour, self.distroseries[1].name, seed, [])

        self.script.process(seed_bases=["file://%s" % self.seeddir])
        self.assertFalse(os.path.exists(os.path.join(
            self.script.miscroot,
            "more-extra.override.%s.main" % self.distroseries[0].name)))
        self.assertTrue(os.path.exists(os.path.join(
            self.script.miscroot,
            "more-extra.override.%s.main" % self.distroseries[1].name)))

    def test_process_removes_only_stale_files(self):
        # The script removes only stale germinate output files.
        flavour = factory.getUniqueString()
        self.setUpDistroAndScript(extra_args=[flavour])
        series_name = self.distroseries[0].name
        self.setUpComponent()
        das = self.distroseries[0].makeDistroArchSeries()
        arch = das.architecture_tag
        self.makeIndexFiles(self.script, self.distroseries[0])

        seed_old = factory.getUniqueString()
        seed_new = factory.getUniqueString()
        self.makeSeedStructure(flavour, series_name, [seed_old])
        self.makeSeed(flavour, series_name, seed_old, [])
        self.script.process(seed_bases=["file://%s" % self.seeddir])
        output = partial(
            self.script.composeOutputPath, flavour, series_name, arch)
        self.assertTrue(os.path.exists(output(seed_old)))
        self.makeSeedStructure(flavour, series_name, [seed_new])
        self.makeSeed(flavour, series_name, seed_new, [])
        self.script.process(seed_bases=["file://%s" % self.seeddir])
        self.assertTrue(os.path.exists(os.path.join(self.script.log_file)))
        self.assertTrue(os.path.exists(output("structure")))
        self.assertTrue(os.path.exists(output("all")))
        self.assertTrue(os.path.exists(output("all.sources")))
        self.assertTrue(os.path.exists(output(seed_new)))
        self.assertFalse(os.path.exists(output(seed_old)))

    def test_process_skips_disabled_distroarchseries(self):
        # The script does not generate overrides for disabled DistroArchSeries.
        flavour = factory.getUniqueString()
        self.setUpDistroAndScript(extra_args=[flavour])
        das = self.distroseries[0].makeDistroArchSeries()
        self.distroseries[0].makeDistroArchSeries(enabled=False)
        self.script.generateExtraOverrides = mock.MagicMock()
        self.script.process()
        self.assertEqual(1, self.script.generateExtraOverrides.call_count)
        self.assertEqual(
            [das.architecture_tag],
            self.script.generateExtraOverrides.call_args[0][2])

    def test_main(self):
        # If run end-to-end, the script generates override files containing
        # output for all architectures, and sends germinate's log output to
        # a file.
        flavour = factory.getUniqueString()
        self.setUpDistroAndScript(extra_args=[flavour])
        series_name = self.distroseries[0].name
        component = self.setUpComponent()
        das_one = self.distroseries[0].makeDistroArchSeries()
        das_two = self.distroseries[0].makeDistroArchSeries()
        package = self.makePackage(component, [das_one, das_two])
        self.makeIndexFiles(self.script, self.distroseries[0])

        seed = factory.getUniqueString()
        self.makeSeedStructure(flavour, series_name, [seed])
        self.makeSeed(
            flavour, series_name, seed, [package],
            headers=["Task-Description: task"])

        self.script.process(seed_bases=["file://%s" % self.seeddir])
        override_path = os.path.join(
            self.script.miscroot, "more-extra.override.%s.main" % series_name)
        expected_overrides = [
            "%s/%s  Task  %s" % (package, das_one.architecture_tag, seed),
            "%s/%s  Task  %s" % (package, das_two.architecture_tag, seed),
            ]
        self.assertCountEqual(
            expected_overrides, file_contents(override_path).splitlines())

        log_file = os.path.join(self.script.germinateroot, "germinate.output")
        self.assertIn("Downloading file://", file_contents(log_file))

    def test_run_script_help(self):
        # The script will run stand-alone and generate help output.
        retval, out, err = run_script(
            "scripts/generate-extra-overrides.py", ["-h"])
        self.assertEqual(0, retval)
        self.assertEqual("", err)
