# Copyright 2014 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""Test for the `get-supported-series` script."""

__metaclass__ = type

import logging
from optparse import OptionValueError
try:
    from unittest import mock
except ImportError:
    import mock

from scripts.base import ScriptFailure
from scripts.get_supported_series import GetSupportedSeries
from tests.testing import (
    factory,
    run_script,
    TestCase,
    )


class MockDistroSeries(mock.MagicMock):
    def __init__(self, distribution, name=None, status="Active Development"):
        super(MockDistroSeries, self).__init__()
        if name is None:
            name = factory.getUniqueString(prefix="distroseries")
        self.name = name
        self.distribution = distribution
        self.status = status


class MockDistribution(mock.MagicMock):
    def __init__(self, name):
        super(MockDistribution, self).__init__()
        self.name = name
        self.series = []

    def makeDistroSeries(self, *args, **kwargs):
        series = MockDistroSeries(distribution=self, *args, **kwargs)
        self.series.append(series)
        return series


class MockLaunchpad(mock.MagicMock):
    def __init__(self):
        super(MockLaunchpad, self).__init__(name="Launchpad")
        self.distributions = {}

    def makeDistribution(self, name=None):
        if name is None:
            name = factory.getUniqueString()
        distribution = MockDistribution(name)
        self.distributions[name] = distribution
        return distribution


class TestGetSupportedSeries(TestCase):
    """Tests for the `GetSupportedSeries` script."""

    def setUp(self):
        super(TestGetSupportedSeries, self).setUp()
        self.launchpad = MockLaunchpad()

    def makeScript(self, distribution, extra_args=None):
        """Create a script for testing."""
        test_args = []
        if distribution is not None:
            test_args.extend(["-d", distribution.name])
        if extra_args is not None:
            test_args.extend(extra_args)
        script = GetSupportedSeries(test_args=test_args)
        for handler in list(script.logger.handlers):
            script.logger.removeHandler(handler)
        script.logger.addHandler(logging.NullHandler())
        script.login = lambda: self.launchpad
        return script

    def test_requires_distro(self):
        # The --distribution or -d argument is mandatory.
        script = self.makeScript(None)
        self.assertRaises(OptionValueError, script.setUp)

    def test_requires_real_distro(self):
        # An incorrect distribution name is flagged as an invalid option
        # value.
        script = self.makeScript(
            None, extra_args=["-d", factory.getUniqueString()])
        self.assertRaises(OptionValueError, script.setUp)

    def test_looks_up_distro(self):
        # The script looks up and keeps the distribution named on the
        # command line.
        distro = self.launchpad.makeDistribution()
        script = self.makeScript(distro)
        script.setUp()
        self.assertEqual(distro, script.distribution)

    def test_requires_series(self):
        # The script fails if the named distribution has no series.
        distro = self.launchpad.makeDistribution()
        script = self.makeScript(distro)
        self.assertRaises(ScriptFailure, script.getSupportedSeries)

    def test_returns_supported_series(self):
        # The script returns a sensible set of supported series if available.
        distro = self.launchpad.makeDistribution()
        distro.makeDistroSeries(name="future", status="Future")
        distro.makeDistroSeries(name="next", status="Pre-release Freeze")
        distro.makeDistroSeries(name="utopic")
        distro.makeDistroSeries(name="trusty", status="Current Stable Release")
        distro.makeDistroSeries(name="saucy", status="Supported")
        distro.makeDistroSeries(name="precise", status="Supported")
        distro.makeDistroSeries(name="warty", status="Obsolete")
        script = self.makeScript(distro)
        self.assertEqual(
            ["next", "utopic", "trusty", "saucy", "precise"],
            script.getSupportedSeries())

    def test_run_script_help(self):
        # The script will run stand-alone and generate help output.
        retval, out, err = run_script(
            "scripts/get-supported-series.py", ["-h"])
        self.assertEqual(0, retval)
        self.assertEqual("", err)
