# Copyright 2011-2014 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""Test utilities.

Many of these are taken more or less directly from Launchpad.
"""

__metaclass__ = type

__all__ = [
    "ensure_directory_exists",
    "factory",
    "file_contents",
    "open_for_writing",
    "run_script",
    "TestCase",
    "write_file",
    ]

import errno
from itertools import count
import logging
import os
import shutil
import subprocess
import sys
import tempfile
import unittest


def ensure_directory_exists(directory, mode=0o777):
    """Create 'directory' if it doesn't exist.

    :return: True if the directory had to be created, False otherwise.
    """
    try:
        os.makedirs(directory, mode=mode)
    except OSError as e:
        if e.errno == errno.EEXIST:
            return False
        raise
    return True


def open_for_writing(filename, mode, dirmode=0o777):
    """Open 'filename' for writing, creating directories if necessary.

    :param filename: The path of the file to open.
    :param mode: The mode to open the filename with. Should be 'w', 'a' or
        something similar. See ``open`` for more details. If you pass in
        a read-only mode (e.g. 'r'), then we'll just accept that and return
        a read-only file-like object.
    :param dirmode: The mode to use to create directories, if necessary.
    :return: A file-like object that can be used to write to 'filename'.
    """
    try:
        return open(filename, mode)
    except IOError as e:
        if e.errno == errno.ENOENT:
            os.makedirs(os.path.dirname(filename), mode=dirmode)
            return open(filename, mode)


def write_file(path, content):
    with open_for_writing(path, 'w') as f:
        f.write(content)


def run_script(script_relpath, args, expect_returncode=0):
    """Run a script for testing purposes.

    :param script_relpath: The relative path to the script, from the tree
        root.
    :param args: Arguments to provide to the script.
    :param expect_returncode: The return code expected.  If a different value
        is returned, and exception will be raised.
    """
    script = os.path.join(sys.path[0], script_relpath)
    args = [script] + args
    process = subprocess.Popen(
        args, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
        universal_newlines=True)
    stdout, stderr = process.communicate()
    if process.returncode != expect_returncode:
        raise AssertionError('Failed:\n%s\n%s' % (stdout, stderr))
    return (process.returncode, stdout, stderr)


def file_contents(path):
    """Return the contents of the file at path."""
    with open(path) as handle:
        return handle.read()


class ObjectFactory:

    _unique_int_counter = count(100000)

    def getUniqueEmailAddress(self):
        return "%s@example.com" % self.getUniqueString("email")

    def getUniqueInteger(self):
        return next(self._unique_int_counter)

    def getUniqueString(self, prefix=None):
        if prefix is None:
            prefix = "unique"
        return "%s-%s" % (prefix, self.getUniqueInteger())


factory = ObjectFactory()


class TestCase(unittest.TestCase):

    def setUp(self):
        super(TestCase, self).setUp()
        self.save_env = dict(os.environ)

    def makeTemporaryDirectory(self):
        """Create a temporary directory, and return its path."""
        temp_dir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, temp_dir)
        return temp_dir

    def useTempDir(self):
        """Use a temporary directory for this test."""
        tempdir = self.makeTemporaryDirectory()
        cwd = os.getcwd()
        os.chdir(tempdir)
        self.addCleanup(os.chdir, cwd)
        return tempdir

    def resetEnvironment(self):
        for key in set(os.environ.keys()) - set(self.save_env.keys()):
            del os.environ[key]
        for key, value in os.environ.items():
            if value != self.save_env[key]:
                os.environ[key] = self.save_env[key]
        for key in set(self.save_env.keys()) - set(os.environ.keys()):
            os.environ[key] = self.save_env[key]

    def resetLogging(self):
        # Remove all handlers from non-root loggers, and remove the loggers
        # too.
        loggerDict = logging.Logger.manager.loggerDict
        for name, logger in list(loggerDict.items()):
            if not isinstance(logger, logging.PlaceHolder):
                for handler in list(logger.handlers):
                    logger.removeHandler(handler)
            del loggerDict[name]

        # Remove all handlers from the root logger
        root = logging.getLogger('')
        for handler in root.handlers:
            root.removeHandler(handler)

        # Set the root logger's log level back to the default level:
        # WARNING.
        root.setLevel(logging.WARNING)

        # Clean out the guts of the logging module.  We don't want handlers
        # that have already been closed hanging around for the atexit
        # handler to barf on, for example.
        del logging._handlerList[:]
        logging._handlers.clear()

    def tearDown(self):
        self.resetEnvironment()
        self.resetLogging()

    # Monkey-patch for Python 2/3 compatibility.
    if not hasattr(unittest.TestCase, 'assertCountEqual'):
        assertCountEqual = unittest.TestCase.assertItemsEqual
